import struct

from aim.ext.transport.message_utils import ResourceObject
from aim.ext.transport.remote_resource import RemoteResourceAutoClean
from aim.storage.treeview import TreeView
from aim.storage.arrayview import TreeArrayView
from aim.storage.types import AimObject, AimObjectKey, AimObjectPath

from typing import TYPE_CHECKING, Iterator, Tuple, Union

if TYPE_CHECKING:
    from aim.ext.transport.client import Client


class ProxyTreeAutoClean(RemoteResourceAutoClean):
    PRIORITY = 60


class ProxyTree(TreeView):
    def __init__(self, client: 'Client',
                 name: str,
                 sub: str,
                 read_only: bool,
                 from_union: bool = False,
                 index=False,
                 timeout=None):
        self._resources: ProxyTreeAutoClean = None

        self._rpc_client = client
        read_only = struct.pack('?', read_only)
        from_union = struct.pack('?', from_union)
        index = struct.pack('?', index)
        timeout = struct.pack('I', timeout or 0)

        args = (name.encode(), sub.encode(), read_only, from_union, index, timeout)
        handler = self._rpc_client.get_resource_handler('TreeView', args=args)

        self._resources = ProxyTreeAutoClean(self)
        self._resources.rpc_client = client
        self._resources.handler = handler
        self._handler = handler

    def preload(self):
        self._rpc_client.run_instruction(self._handler, 'preload')

    def view(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        resolve: bool = False
    ):
        subtree = SubtreeView(self, path)
        if not resolve:
            return subtree
        # TODO [AT, MV] handle resolve=True
        # make an rpc call to get underlying object type
        # and construct CustomObject if needed
        return None

    def make_array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ):
        self._rpc_client.run_instruction(self._handler, 'make_array', (path,))

    def collect(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        strict: bool = True
    ) -> AimObject:
        return self._rpc_client.run_instruction(self._handler, 'collect', (path, strict))

    def __delitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ):
        self._rpc_client.run_instruction(self._handler, '__delitem__', (path,))

    def __setitem__(
            self,
            path: Union[AimObjectKey, AimObjectPath],
            value: AimObject
    ):
        self._rpc_client.run_instruction(self._handler, '__setitem__', (path, value))

    def keys(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = None
    ) -> Iterator[Union[AimObjectPath, AimObjectKey]]:
        return self._rpc_client.run_instruction(self._handler, 'keys', (path, level))

    def items(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Iterator[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        return self._rpc_client.run_instruction(self._handler, 'items', (path,))

    def iterlevel(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = 1
    ) -> Iterator[Tuple[
        AimObjectPath,
        AimObject
    ]]:
        return self._rpc_client.run_instruction(self._handler, 'iterlevel', (path, level))

    def array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> TreeArrayView:
        return TreeArrayView(self.subtree(path))

    def first(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Tuple[AimObjectKey, AimObject]:
        return self._rpc_client.run_instruction(self._handler, 'first', (path,))

    def last(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Tuple[AimObjectKey, AimObject]:
        return self._rpc_client.run_instruction(self._handler, 'last', (path,))

    def finalize(
        self,
        *,
        index: 'ProxyTree'
    ):
        self._rpc_client.run_instruction(self._handler, 'finalize', (ResourceObject(index._handler),))


class SubtreeView(TreeView):
    def __init__(self, tree: TreeView, path: Union[AimObjectKey, AimObjectPath]):
        self.tree = tree
        if path == Ellipsis:
            path = ()
        if not isinstance(path, (tuple, list)):
            path = (path,)
        self.prefix = path

    def absolute_path(self, path):
        if path == Ellipsis:
            path = ()
        if not isinstance(path, (tuple, list)):
            path = (path,)
        return self.prefix + path

    def preload(self):
        self.tree.preload()

    def view(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        resolve: bool = False
    ):
        return self.tree.view(self.absolute_path(path), resolve)

    def make_array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ):
        self.tree.make_array(self.absolute_path(path))

    def collect(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        strict: bool = True
    ) -> AimObject:
        return self.tree.collect(self.absolute_path(path), strict)

    def __delitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ):
        del self.tree[self.absolute_path(path)]

    def __setitem__(
            self,
            path: Union[AimObjectKey, AimObjectPath],
            value: AimObject
    ):
        self.tree[self.absolute_path(path)] = value

    def keys(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = None
    ) -> Iterator[Union[AimObjectPath, AimObjectKey]]:
        return self.tree.keys(self.absolute_path(path))

    def items(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Iterator[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        return self.tree.items(self.absolute_path(path))

    def iterlevel(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = 1
    ) -> Iterator[Tuple[
        AimObjectPath,
        AimObject
    ]]:
        return self.tree.iterlevel(self.absolute_path(path), level)

    def array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> TreeArrayView:
        return TreeArrayView(self.subtree(path))

    def first(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Tuple[AimObjectKey, AimObject]:
        return self.tree.first(self.absolute_path(path))

    def last(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Tuple[AimObjectKey, AimObject]:
        return self.tree.last(self.absolute_path(path))



import os
import uuid
from copy import deepcopy
import grpc
import aim.ext.transport.remote_tracking_pb2 as rpc_messages
import aim.ext.transport.remote_tracking_pb2_grpc as remote_tracking_pb2_grpc

from aim.ext.transport.message_utils import pack_stream, unpack_stream, raise_exception
from aim.ext.transport.config import AIM_CLIENT_SSL_CERTIFICATES_FILE
from aim.storage.treeutils import encode_tree, decode_tree


class Client:
    def __init__(self, remote_path: str):
        self._id = str(uuid.uuid4())

        ssl_certfile = os.getenv(AIM_CLIENT_SSL_CERTIFICATES_FILE)
        if ssl_certfile:
            with open(ssl_certfile, 'rb') as f:
                root_certificates = grpc.ssl_channel_credentials(f.read())
            self._remote_channel = grpc.secure_channel(remote_path, root_certificates)
        else:
            self._remote_channel = grpc.insecure_channel(remote_path)

        self._remote_stub = remote_tracking_pb2_grpc.RemoteTrackingServiceStub(self._remote_channel)

    def get_resource_handler(self, resource_type, args=()):
        request = rpc_messages.ResourceRequest(
            resource_type=resource_type,
            client_uri=self.uri,
            args=args
        )
        response = self.remote.get_resource(request)
        if response.status == rpc_messages.ResourceResponse.Status.ERROR:
            raise_exception(response.exception)
        return response.handler

    def release_resource(self, resource_handler):
        request = rpc_messages.ReleaseResourceRequest(
            handler=resource_handler,
            client_uri=self.uri
        )
        response = self.remote.release_resource(request)
        if response.status == rpc_messages.ReleaseResourceResponse.Status.ERROR:
            raise_exception(response.exception)

    def run_instruction(self, resource, method, args=()):
        args = deepcopy(args)

        def message_stream_generator():
            header = rpc_messages.InstructionRequest(
                header=rpc_messages.RequestHeader(
                    version='0.1',
                    handler=resource,
                    client_uri=self.uri,
                    method_name=method
                )
            )
            yield header

            stream = pack_stream(encode_tree(args))
            for chunk in stream:
                yield rpc_messages.InstructionRequest(message=chunk)

        resp = self.remote.run_instruction(message_stream_generator())
        status_msg = next(resp)
        assert status_msg.WhichOneof('instruction') == 'header'
        if status_msg.header.status == rpc_messages.ResponseHeader.Status.ERROR:
            raise_exception(status_msg.header.exception)
        return decode_tree(unpack_stream(resp))

    @property
    def remote(self):  # access to low-level interface
        return self._remote_stub

    @property
    def uri(self):
        return self._id



import json
from typing import List
from typing import TYPE_CHECKING

from aim.sdk.sequence import Sequence
from aim.sdk.sequence_collection import SequenceCollection
from aim.storage.treeutils import encode_tree

from aim.storage.context import Context
from aim.sdk.objects import Figure
from aim.sdk.run import Run
from aim.sdk.uri_service import URIService, generate_resource_path

from aim.web.api.runs.utils import (
    collect_run_streamable_data,
    IndexRange,
    get_run_props,
)
from aim.web.api.runs.pydantic_models import TraceBase

if TYPE_CHECKING:
    from aim.sdk import Repo


def preparer(obj: Figure, trace, step, decode=False):
    assert isinstance(obj, Figure)

    if decode:
        return json.loads(obj.json())
    resource_path = generate_resource_path(trace.values.tree.container, (step, 'data'))
    return {
        'blob_uri': URIService.generate_uri(
            trace.run.repo, trace.run.hash, 'seqs', resource_path
        )
    }


def get_record_and_index_range(
        traces: SequenceCollection, trace_cache: dict
) -> IndexRange:
    rec_start = None
    rec_stop = -1
    for run_trace_collection in traces.iter_runs():
        run = run_trace_collection.run
        run_traces = []
        for trace in run_trace_collection.iter():
            run_traces.append(trace)
            rec_start = (
                min(trace.first_step(), rec_start) if rec_start else trace.first_step()
            )
            rec_stop = max(trace.last_step(), rec_stop)
        if run_traces:
            trace_cache[run.hash] = {'run': run, 'traces': run_traces}
    return IndexRange(rec_start, rec_stop + 1)


def get_trace_info(trace: Sequence, rec_slice: slice, rec_density: int) -> dict:
    steps = []
    values = []
    steps_vals = trace.values.items_in_range(
        rec_slice.start, rec_slice.stop, rec_density
    )
    for step, val in steps_vals:
        steps.append(step)
        values.append(preparer(val, trace, step))
    return {
        'name': trace.name,
        'context': trace.context.to_dict(),
        'values': values,
        'iters': steps,
        'epochs': list(
            trace.epochs.values_in_range(rec_slice.start, rec_slice.stop, rec_density)
        ),
        'timestamps': list(
            trace.timestamps.values_in_range(
                rec_slice.start, rec_slice.stop, rec_density
            )
        ),
    }


async def figure_search_result_streamer(
        traces: SequenceCollection,
        rec_range: IndexRange,
        rec_density: int,
        calc_total_ranges: bool,
):
    record_range_missing = rec_range.start is None or rec_range.stop is None
    run_traces = {}

    trcs_rec_range = IndexRange(None, None)
    if record_range_missing or calc_total_ranges:
        trcs_rec_range = get_record_and_index_range(traces, trace_cache=run_traces)
        if not run_traces:
            return

    rec_start = rec_range.start if rec_range.start is not None else trcs_rec_range.start
    rec_stop = rec_range.stop if rec_range.stop is not None else trcs_rec_range.stop
    rec_step = (rec_stop - rec_start) // rec_density or 1

    rec_slice = slice(rec_start, rec_stop, rec_step)

    def _pack_run_data(run_: Run, traces_: list):
        _rec_range = (
            trcs_rec_range if record_range_missing or calc_total_ranges else rec_range
        )

        run_dict = {
            run_.hash: {
                'ranges': {
                    'record_range': [_rec_range.start, _rec_range.stop],
                    'record_slice': [rec_slice.start, rec_slice.stop, rec_slice.step],
                },
                'params': run_.get(...),
                'traces': traces_,
                'props': get_run_props(run_),
            }
        }
        encoded_tree = encode_tree(run_dict)
        return collect_run_streamable_data(encoded_tree)

    if run_traces:
        for run_info in run_traces.values():
            traces_list = []
            for trace in run_info['traces']:
                traces_list.append(get_trace_info(trace, rec_slice, rec_density))
            yield _pack_run_data(run_info['run'], traces_list)
    else:
        for run_trace_collection in traces.iter_runs():
            traces_list = []
            for trace in run_trace_collection.iter():
                traces_list.append(get_trace_info(trace, rec_slice, rec_density))
            if traces_list:
                yield _pack_run_data(run_trace_collection.run, traces_list)


def figure_batch_result_streamer(uri_batch: List[str], repo: 'Repo'):
    uri_service = URIService(repo=repo)
    batch_iterator = uri_service.request_batch(uri_batch=uri_batch)
    for it in batch_iterator:
        yield collect_run_streamable_data(encode_tree(it))


def requested_figure_object_traces_streamer(
        run: Run, requested_traces: List[TraceBase], rec_range, rec_num: int = 50
) -> List[dict]:
    for requested_trace in requested_traces:
        trace_name = requested_trace.name
        context = Context(requested_trace.context)
        trace = run.get_figure_sequence(name=trace_name, context=context)
        if not trace:
            continue

        record_range_missing = rec_range.start is None or rec_range.stop is None
        if record_range_missing:
            rec_range = IndexRange(trace.first_step(), trace.last_step() + 1)

        steps = []
        values = []
        steps_vals = trace.values.items_in_range(
            rec_range.start, rec_range.stop, rec_num
        )
        for step, val in steps_vals:
            steps.append(step)
            values.append(preparer(val, trace, step, decode=True))

        trace_dict = {
            'name': trace.name,
            'context': trace.context.to_dict(),
            'values': values,
            'iters': steps,
            'record_range': (trace.first_step(), trace.last_step() + 1),
        }
        encoded_tree = encode_tree(trace_dict)
        yield collect_run_streamable_data(encoded_tree)



import os
import struct

from aim.sdk import Repo
from aim.ext.transport.config import AIM_SERVER_MOUNTED_REPO_PATH


def get_tree(name: bytes, sub: bytes, read_only: bytes, from_union: bytes, index: bytes, timeout: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()

    read_only = struct.unpack('?', read_only)[0]
    from_union = struct.unpack('?', from_union)[0]
    index = struct.unpack('?', index)[0]
    timeout = struct.unpack('I', timeout)[0]
    if index:
        return repo._get_index_tree(name.decode(), timeout)
    else:
        return repo.request_tree(name.decode(), sub.decode(), read_only=read_only, from_union=from_union)


def get_structured_run(hash_: bytes, read_only: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()
    read_only = struct.unpack('?', read_only)[0]
    return repo.request_props(hash_.decode(), read_only)



# AIM SERVER
AIM_SERVER_DEFAULT_PORT = 53800
AIM_SERVER_DEFAULT_HOST = '0.0.0.0'
AIM_SERVER_MOUNTED_REPO_PATH = '__AIM_SERVER_MOUNT_REPO_PATH__'

# AIM CLIENT
# The path of the PEM-encoded root certificates. Used for secure channel establishment
AIM_CLIENT_SSL_CERTIFICATES_FILE = '__AIM_CLIENT_SSL_CERTIFICATES_FILE__'



import click

from aim.sdk.repo import Repo, RepoStatus
from aim.sdk.utils import clean_repo_path
from aim.cli.convert.utils import parse_tf_events


@click.group()
@click.option('--repo', required=False, type=click.Path(exists=True,
                                                        file_okay=False,
                                                        dir_okay=True,
                                                        writable=True))
@click.pass_context
def convert(ctx, repo):
    ctx.ensure_object(dict)

    repo_path = clean_repo_path(repo) or Repo.default_repo_path()
    repo_status = Repo.check_repo_status(repo_path)

    repo_inst = Repo.from_path(repo_path)
    if repo_status == RepoStatus.PATCH_REQUIRED:
        repo_inst.structured_db.run_upgrades()

    ctx.obj['repo_inst'] = repo_inst


@convert.command(name='tf')
@click.pass_context
@click.option('--logdir', required=True, type=click.Path(exists=True,
                                                         readable=True,
                                                         dir_okay=True,
                                                         resolve_path=True))
@click.option('--flat', '-f', required=False, is_flag=True, default=False)
def convert_tensorflow(ctx, logdir, flat):
    repo_inst = ctx.obj['repo_inst']
    parse_tf_events(logdir, repo_inst, flat)



import json
import os

import click

from aim import Image, Run


def parse_tf_events(tf_logs, repo_inst, flat=False):
    """
    This function scans and collects records from TF event files.

    Creates and uses cache file "tf_logs_cache" in the repo dir
    to track previously processed files and values

    For more info please refer to this guide: docs/source/guides/integrations/basic_aim_tensorflow_event_conversion.md
    """

    try:
        # This import statement takes long to complete
        import tensorflow as tf
        from tensorflow.python.summary.summary_iterator import summary_iterator
    except ModuleNotFoundError:
        click.echo(
            click.style(
                'Could not process TF events - failed to import tensorflow module.', fg='red'
            )
        )
        return

    supported_plugins = ('images', 'scalars')
    unsupported_plugin_noticed = False
    tf_logs_cache_path = os.path.join(repo_inst.path, 'tf_logs_cache')

    try:
        with open(tf_logs_cache_path) as FS:
            tf_logs_cache = json.load(FS)
    except Exception:
        tf_logs_cache = {}

    def get_parent(current_path, level=0):
        # level 0 is the direct parent directory
        if level <= 0:
            return os.path.dirname(current_path)
        elif current_path in ('', '.', '/'):
            return current_path
        return get_parent(os.path.dirname(current_path), level - 1)

    tf_logs = os.path.abspath(tf_logs)
    run_dir_candidates = set()
    for root, dirs, files in os.walk(tf_logs):
        for file in files:
            if not file.startswith('events.out.tfevents'):
                continue

            file_path = os.path.abspath(os.path.join(root, file))
            run_dir = get_parent(file_path)

            if not run_dir.startswith(tf_logs):
                # it's outside tf_logs
                continue

            run_dir_candidates.add(run_dir)

    def get_level(current_path):
        level = -1
        while current_path.startswith(tf_logs):
            current_path, _ = os.path.split(current_path)
            level += 1
        return level

    run_dir_candidates = sorted(run_dir_candidates, key=get_level, reverse=True)
    run_dir_candidates_filtered = set()
    run_dir_ignored = set()
    groups = set()

    for run_dir in run_dir_candidates:
        if run_dir in run_dir_candidates_filtered:
            # already tagged as a run dir
            continue

        if run_dir in groups:
            # run dir which has other run dirs inside, so we skip it
            run_dir_ignored.add(run_dir)
            continue

        depth = get_level(run_dir)
        if depth >= 2:
            if flat:
                run_group_dir = get_parent(run_dir, 0)
                new_run_dir = run_dir
            else:
                run_group_dir = get_parent(run_dir, 1)
                new_run_dir = get_parent(run_dir, 0)
                if new_run_dir in groups:
                    new_run_dir = run_dir
            groups.add(run_group_dir)
        elif depth == 1:
            new_run_dir = run_dir
        else:
            continue
        run_dir_candidates_filtered.add(new_run_dir)

    if run_dir_ignored:
        click.echo('WARN: Found directory entries with unorganized even files!\n'
                   'Please read the preparation instructions to properly process these files.\n'
                   'Event files in the following directories will be ignored:', err=True)
        for c, r in enumerate(run_dir_ignored, start=1):
            click.echo(f'{c}: {r}', err=True)

    for path in run_dir_candidates_filtered:
        click.echo(f'Converting TensorFlow events in {path}')

        events = {}
        for root, dirs, files in os.walk(path):
            for file in files:
                if 'events.out.tfevents' not in file:
                    continue
                file_path = os.path.join(root, file)
                if file_path == os.path.join(path, file):
                    entry = None
                else:
                    entry = os.path.basename(os.path.dirname(file_path))
                events[file_path] = {
                    'context': {
                        'entry': entry
                    }
                }

        if path not in tf_logs_cache:
            tf_logs_cache[path] = {}

        run_cache = tf_logs_cache[path]
        if run_cache:
            run = Run(
                repo=repo_inst,
                system_tracking_interval=None,
                run_hash=run_cache['run_hash'],
            )
        else:
            run = Run(
                repo=repo_inst,
                system_tracking_interval=None,
            )
            run['tf_logdir'] = path
            run_cache.update({
                'run_hash': run.hash,
                'events': {},
            })
        run_tf_events = run_cache['events']

        events_to_process = []
        for event in events:
            last_modified_at = os.path.getmtime(event)
            try:
                assert last_modified_at == run_tf_events[event]['last_modified_at']
            except (KeyError, AssertionError):
                # Something has changed or hasn't been processed before
                events_to_process.append(event)
                try:
                    run_tf_events[event]['last_modified_at'] = last_modified_at
                except KeyError:
                    # Completely new event
                    run_tf_events[event] = {
                        'last_modified_at': last_modified_at,
                        'values': {},
                    }

        for count, event_file in enumerate(events_to_process, start=1):
            click.echo(f'({count}/{len(events_to_process)}) Parsing TF event: {os.path.basename(event_file)}')
            run_tf_event = run_tf_events[event_file]
            event_context = events[event_file]['context']
            for event in summary_iterator(event_file):
                timestamp = event.wall_time
                step = event.step
                for value in event.summary.value:
                    tag = value.tag
                    plugin_name = value.metadata.plugin_data.plugin_name
                    value_id = f'{tag}_{plugin_name}'
                    if value_id in run_tf_event['values']:
                        if run_tf_event['values'][value_id]['timestamp'] >= timestamp:
                            # prevent previously tracked data from re-tracking upon file update
                            continue

                    if plugin_name not in supported_plugins:
                        if not unsupported_plugin_noticed:
                            click.echo(
                                'Found unsupported plugin type in the TF events. '
                                'Data for these wont be processed. '
                                'Supported plugin types are: {}'.format(', '.join(supported_plugins)),
                                err=True
                            )
                            unsupported_plugin_noticed = True
                        continue

                    if plugin_name == 'images':
                        tf_tensors = value.tensor.string_val[2:]
                        track_val = [
                            Image(tf.image.decode_image(tf_tensor).numpy()) for tf_tensor in tf_tensors
                        ]
                        if len(track_val) == 1:
                            track_val = track_val[0]
                    else:
                        d_type = tf.dtypes.DType(value.tensor.dtype)
                        decoded = tf.io.decode_raw(value.tensor.tensor_content, d_type)
                        track_val = float(decoded)

                    run_tf_event['values'][value_id] = {
                        'step': step,
                        'timestamp': timestamp
                    }
                    run._track_impl(track_val, timestamp, tag, step, context=event_context)

    # refresh cache
    with open(tf_logs_cache_path, 'w') as FS:
        json.dump(tf_logs_cache, FS)

    click.echo('TF logs conversion complete!')



import numpy as np

from typing import Tuple

from aim.storage.object import CustomObject
from aim.storage.types import BLOB


@CustomObject.alias('aim.distribution')
class Distribution(CustomObject):
    """Distribution object used to store distribution objects in Aim repository.

    Args:
         distribution (:obj:): array-like object used to construct `aim.Distribution`.
         bin_count (:obj:`int`, optional): Optional distribution bin count. 64 by default, max 512.
    """

    AIM_NAME = 'aim.distribution'

    def __init__(self, distribution, bin_count=64):
        super().__init__()

        if not isinstance(bin_count, int):
            raise TypeError('`bin_count` must be an integer.')
        if 1 > bin_count > 512:
            raise ValueError('Supported range for `bin_count` is [1, 512].')
        self.storage['bin_count'] = bin_count

        # convert to np.histogram
        try:
            np_histogram = np.histogram(distribution, bins=bin_count)
        except TypeError:
            raise TypeError(f'Cannot convert to aim.Distribution. Unsupported type {type(distribution)}.')
        self._from_np_histogram(np_histogram)

    @property
    def bin_count(self):
        """Stored distribution bin count

            :getter: Returns distribution bin_count.
            :type: string
        """
        return self.storage['bin_count']

    @property
    def range(self):
        """Stored distribution range

            :getter: Returns distribution range.
            :type: List
        """
        return self.storage['range']

    @property
    def weights(self):
        """Stored distribution weights

            :getter: Returns distribution weights as `np.array`.
            :type: np.ndarray
        """
        return np.frombuffer(self.storage['data'].load(),
                             dtype=self.storage['dtype'],
                             count=self.storage['bin_count'])

    @property
    def ranges(self):
        """Stored distribution ranges

            :getter: Returns distribution ranges as `np.array`.
            :type: np.ndarray
        """
        assert (len(self.range) == 2)
        return np.linspace(self.range[0], self.range[1], num=self.bin_count)

    def json(self):
        """Dump distribution metadata to a dict"""
        return {
            'bin_count': self.bin_count,
            'range': self.range,
        }

    def _from_np_histogram(self, np_histogram: Tuple[np.ndarray, np.ndarray]):
        assert isinstance(np_histogram[0], np.ndarray)
        assert isinstance(np_histogram[1], np.ndarray)

        self.storage['data'] = BLOB(data=np_histogram[0].tobytes())
        self.storage['dtype'] = str(np_histogram[0].dtype)
        self.storage['range'] = [np_histogram[1][0].item(), np_histogram[1][-1].item()]

    def to_np_histogram(self):
        """Return `np.histogram` compatible format of the distribution"""
        return self.weights, self.ranges
